﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using XorShift;

namespace DungeonGenerator131202
{
	public sealed class Map
	{
		private readonly Xor128 rnd;
		private readonly uint seedX;
		private readonly uint seedY;
		private readonly uint seedZ;
		private readonly uint seedW;
		private readonly Grid[][] grids;

		public uint SeedX
		{
			get
			{
				return this.seedX;
			}
		}
		public uint SeedY
		{
			get
			{
				return this.seedY;
			}
		}
		public uint SeedZ
		{
			get
			{
				return this.seedZ;
			}
		}
		public uint SeedW
		{
			get
			{
				return this.seedW;
			}
		}
		public Grid[][] Grids
		{
			get
			{
				return this.grids;
			}
		}

		public Map(uint seedX, uint seedY, uint seedZ, uint seedW, int mapWidth, int mapHeight, int minRoomWidth, int maxRoomWidth, int minRoomHeight, int maxRoomHeight, int minRooms, int maxRooms, int noOutletRate, int intersectRate, int maxConnectRoomsLength, int permissibleFail)
		{
			--minRoomWidth;
			--minRoomHeight;
			if ((seedX | seedY | seedZ | seedW) != 0)
			{
				this.rnd = new Xor128(seedX, seedY, seedZ, seedW);
				this.seedX = seedX;
				this.seedY = seedY;
				this.seedZ = seedZ;
				this.seedW = seedW;
			}
			else
			{
				this.rnd = new Xor128(ref this.seedX, ref this.seedY, ref this.seedZ, ref this.seedW);
			}
			this.grids = new Grid[mapHeight][];
			for (int y = 0; y < mapHeight; ++y)
			{
				this.grids[y] = new Grid[mapWidth];
			}
			int rooms = this.rnd.Next(minRooms, maxRooms);
			List<Rectangle> roomList = new List<Rectangle>();
			for (int i = 0, fail = 0, width, height; i < rooms; ++i)
			{
				bool noOutlet;
			Start:
				if (noOutletRate >= 100 || (noOutletRate > 0 && this.rnd.Next(100) < noOutletRate))
				{
					width = height = 0;
					noOutlet = true;
				}
				else
				{
					width = this.rnd.Next(minRoomWidth, maxRoomWidth);
					height = this.rnd.Next(minRoomWidth, maxRoomWidth);
					noOutlet = false;
				}
				Rectangle r1 = new Rectangle(this.rnd.Next(mapWidth - width), this.rnd.Next(mapHeight - height), width, height);
				for (int y = r1.Top; y <= r1.Bottom; ++y)
				{
					for (int x = r1.Left; x <= r1.Right; ++x)
					{
						if (this.grids[y][x] != Grid.Zero)
						{
							if (++fail > permissibleFail)
							{
								return;
							}
							goto Start;
						}
					}
				}
				if (i == 0)
				{
					this.SetRoomGrids(r1, noOutlet);
					roomList.Add(r1);
					continue;
				}
				Rectangle r2;
				Point start, relay, goal;
				for (int j = 0, index = i == 1 ? 0 : this.rnd.Next(i); ; )
				{
					r2 = roomList[index];
					for (int k = 0, side = this.rnd.Next(4); k < 4; ++k)
					{
						switch (side)
						{
							case 0:	// Top
								if (r1.Top <= r2.Top || r1.Top - r2.Bottom > maxConnectRoomsLength)
								{
									break;
								}
								for (int l = 0, direction = this.rnd.Next(3); l < 3; ++l)
								{
									switch (direction)
									{
										case 0:	// North West
											if (r1.Right <= r2.Right || r1.Left - r2.Right > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(this.rnd.Next(Math.Max(r1.Left, r2.Left), r1.Right + 1), r1.Top);
											if (start.X > r2.Right)
											{
												goal = new Point(r2.Right, this.rnd.Next(r2.Top, Math.Min(r2.Bottom + 1, r1.Top)));
												relay = new Point(start.X, goal.Y);
												goto CreatePassage;
											}
											goal = new Point(start.X, r2.Bottom);
											goto CreateOneLinePassage;
										case 1:	// North
											if ((r1.Left > r2.Left && r1.Right > r2.Right) || (r2.Left > r1.Left && r2.Right > r1.Right))
											{
												break;
											}
											start = new Point(this.rnd.Next(Math.Max(r1.Left, r2.Left), Math.Min(r1.Right, r2.Right) + 1), r1.Top);
											goal = new Point(start.X, r2.Bottom);
											goto CreateOneLinePassage;
										case 2:	// North East
											if (r2.Left <= r1.Left || r2.Left - r1.Right > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(this.rnd.Next(r1.Left, Math.Min(r1.Right, r2.Right) + 1), r1.Top);
											if (r2.Left > start.X)
											{
												goal = new Point(r2.Left, this.rnd.Next(r2.Top, Math.Min(r2.Bottom + 1, r1.Top)));
												relay = new Point(start.X, goal.Y);
												goto CreatePassage;
											}
											goal = new Point(start.X, r2.Bottom);
											goto CreateOneLinePassage;
									}
									if (++direction == 3)
									{
										direction = 0;
									}
								}
								break;
							case 1:	// Bottom
								if (r2.Bottom <= r1.Bottom || r2.Top - r1.Bottom > maxConnectRoomsLength)
								{
									break;
								}
								for (int l = 0, direction = this.rnd.Next(3); l < 3; ++l)
								{
									switch (direction)
									{
										case 0:	// South West
											if (r1.Right <= r2.Right || r1.Left - r2.Right > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(this.rnd.Next(Math.Max(r1.Left, r2.Left), r1.Right + 1), r1.Bottom);
											if (start.X > r2.Right)
											{
												goal = new Point(r2.Right, this.rnd.Next(Math.Max(r2.Top, r1.Bottom + 1), r2.Bottom));
												relay = new Point(start.X, goal.Y);
												goto CreatePassage;
											}
											goal = new Point(start.X, r2.Top);
											goto CreateOneLinePassage;
										case 1:	// South
											if ((r1.Left > r2.Left && r1.Right > r2.Right) || (r2.Left > r1.Left && r2.Right > r1.Right))
											{
												break;
											}
											start = new Point(this.rnd.Next(Math.Max(r1.Left, r2.Left), Math.Min(r1.Right, r2.Right) + 1), r1.Bottom);
											goal = new Point(start.X, r2.Top);
											goto CreateOneLinePassage;
										case 2:	// South East
											if (r2.Left <= r1.Left || r2.Left - r1.Right > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(this.rnd.Next(r1.Left, Math.Min(r1.Right, r2.Right) + 1), r1.Bottom);
											if (r2.Left > start.X)
											{
												goal = new Point(r2.Left, this.rnd.Next(Math.Max(r2.Top, r1.Bottom + 1), r2.Bottom + 1));
												relay = new Point(start.X, goal.Y);
												goto CreatePassage;
											}
											goal = new Point(start.X, r2.Top);
											goto CreateOneLinePassage;
									}
									if (++direction == 3)
									{
										direction = 0;
									}
								}
								break;
							case 2:	// Left
								if (r1.Left <= r2.Left || r1.Left - r2.Right > maxConnectRoomsLength)
								{
									break;
								}
								for (int l = 0, direction = this.rnd.Next(3); l < 3; ++l)
								{
									switch (direction)
									{
										case 0:	// West North
											if (r1.Bottom <= r2.Bottom || r1.Top - r2.Bottom > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(r1.Left, this.rnd.Next(Math.Max(r1.Top, r2.Top), r1.Bottom + 1));
											if (start.Y > r2.Bottom)
											{
												goal = new Point(this.rnd.Next(r2.Left, Math.Min(r2.Right + 1, r1.Left)), r2.Bottom);
												relay = new Point(goal.X, start.Y);
												goto CreatePassage;
											}
											goal = new Point(r2.Right, start.Y);
											goto CreateOneLinePassage;
										case 1:	// West
											if ((r1.Top > r2.Top && r1.Bottom > r2.Bottom) || (r2.Top > r1.Top && r2.Bottom > r1.Bottom))
											{
												break;
											}
											start = new Point(r1.Left, this.rnd.Next(Math.Max(r1.Top, r2.Top), Math.Min(r1.Bottom, r2.Bottom) + 1));
											goal = new Point(r2.Right, start.Y);
											goto CreateOneLinePassage;
										case 2:	// West South
											if (r2.Top <= r1.Top || r2.Top - r1.Bottom > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(r1.Left, this.rnd.Next(r1.Top, Math.Min(r1.Bottom, r2.Bottom) + 1));
											if (r2.Top > start.Y)
											{
												goal = new Point(this.rnd.Next(r2.Left, Math.Min(r2.Right + 1, r1.Left)), r2.Top);
												relay = new Point(goal.X, start.Y);
												goto CreatePassage;
											}
											goal = new Point(r2.Right, start.Y);
											goto CreateOneLinePassage;
									}
									if (++direction == 3)
									{
										direction = 0;
									}
								}
								break;
							case 3:	// Right
								if (r2.Right <= r1.Right || r2.Left - r1.Right > maxConnectRoomsLength)
								{
									break;
								}
								for (int l = 0, direction = this.rnd.Next(3); l < 3; ++l)
								{
									switch (direction)
									{
										case 0:	// East North
											if (r1.Bottom <= r2.Bottom || r1.Top - r2.Bottom > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(r1.Right, this.rnd.Next(Math.Max(r1.Top, r2.Top), r1.Bottom + 1));
											if (start.Y > r2.Bottom)
											{
												goal = new Point(this.rnd.Next(Math.Max(r2.Left, r1.Right + 1), r2.Right), r2.Bottom);
												relay = new Point(goal.X, start.Y);
												goto CreatePassage;
											}
											goal = new Point(r2.Right, start.Y);
											goto CreateOneLinePassage;
										case 1:	// East
											if ((r1.Top > r2.Top && r1.Bottom > r2.Bottom) || (r2.Top > r1.Top && r2.Bottom > r1.Bottom))
											{
												break;
											}
											start = new Point(r1.Right, this.rnd.Next(Math.Max(r1.Top, r2.Top), Math.Min(r1.Bottom, r2.Bottom) + 1));
											goal = new Point(r2.Left, start.Y);
											goto CreateOneLinePassage;
										case 2:	// East South
											if (r2.Top <= r1.Top || r2.Top - r1.Bottom > maxConnectRoomsLength)
											{
												break;
											}
											start = new Point(r1.Right, this.rnd.Next(r1.Top, Math.Min(r1.Bottom, r2.Bottom) + 1));
											if (r2.Top > start.Y)
											{
												goal = new Point(this.rnd.Next(Math.Max(r2.Left, r1.Right + 1), r2.Right + 1), r2.Top);
												relay = new Point(goal.X, start.Y);
												goto CreatePassage;
											}
											goal = new Point(r2.Right, start.Y);
											goto CreateOneLinePassage;
									}
									if (++direction == 3)
									{
										direction = 0;
									}
								}
								break;
						}
						if (++side == 4)
						{
							side = 0;
						}
					}
					if (++j == i)
					{
						if (++fail > permissibleFail)
						{
							return;
						}
						goto Start;
					}
					else if (++index == i)
					{
						index = 0;
					}
				}
			CreateOneLinePassage:
				relay = new Point(-1, 0);
			CreatePassage:
				this.SetRoomGrids(r1, noOutlet);
				Point[] points = relay.X < 0 ? new Point[] { start, goal } : new Point[] { start, relay, goal };
				for (int j = 1; j < points.Length; ++j)
				{
					int x = points[j - 1].X, y = points[j - 1].Y, x2 = points[j].X, y2 = points[j].Y;
					if (x == x2)
					{
						if (y > y2)
						{
							if (j == 1)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y][x] |= Grid.Passage | Grid.TopGate;
									this.grids[y - 1][x] |= Grid.BottomGate;
								}
								else
								{
									this.grids[y][x] |= Grid.Passage;
									this.grids[y][x] &= ~Grid.TopWall;
								}
							}
							else
							{
								this.grids[y][x] &= ~Grid.TopWall;
							}
							while (--y >= y2)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y + 1][x] |= Grid.TopGate;
									this.grids[y][x] |= Grid.Passage | Grid.BottomGate;
									goto EndCreatePassage;
								}
								else if ((this.grids[y][x] & Grid.Passage) != 0)
								{
									this.grids[y][x] &= ~Grid.BottomWall;
									if (y == y2 || (this.grids[y][x] & Grid.TopWall) != 0 || (intersectRate < 100 && this.rnd.Next(100) >= intersectRate))
									{
										goto EndCreatePassage;
									}
									this.grids[y][x] &= ~Grid.TopWall;
								}
								else
								{
									this.grids[y][x] |= Grid.VerticalPassage;
									if (y == y2)
									{
										this.grids[y][x] |= Grid.TopWall;
									}
								}
							}
						}
						else
						{
							if (j == 1)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y][x] |= Grid.Passage | Grid.BottomGate;
									this.grids[y + 1][x] |= Grid.TopGate;
								}
								else
								{
									this.grids[y][x] |= Grid.Passage;
									this.grids[y][x] &= ~Grid.BottomWall;
								}
							}
							else
							{
								this.grids[y][x] &= ~Grid.BottomWall;
							}
							while (++y <= y2)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y - 1][x] |= Grid.BottomGate;
									this.grids[y][x] |= Grid.Passage | Grid.TopGate;
									goto EndCreatePassage;
								}
								else if ((this.grids[y][x] & Grid.Passage) != 0)
								{
									this.grids[y][x] &= ~Grid.TopWall;
									if (y == y2 || (this.grids[y][x] & Grid.BottomWall) != 0 || (intersectRate < 100 && this.rnd.Next(100) >= intersectRate))
									{
										goto EndCreatePassage;
									}
									this.grids[y][x] &= ~Grid.BottomWall;
								}
								else
								{
									this.grids[y][x] |= Grid.VerticalPassage;
									if (y == y2)
									{
										this.grids[y][x] |= Grid.BottomWall;
									}
								}
							}
						}
					}
					else
					{
						if (x > x2)
						{
							if (j == 1)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y][x] |= Grid.Passage | Grid.LeftGate;
									this.grids[y][x - 1] |= Grid.RightGate;
								}
								else
								{
									this.grids[y][x] |= Grid.Passage;
									this.grids[y][x] &= ~Grid.LeftWall;
								}
							}
							else
							{
								this.grids[y][x] &= ~Grid.LeftWall;
							}
							while (--x >= x2)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y][x + 1] |= Grid.LeftGate;
									this.grids[y][x] |= Grid.Passage | Grid.RightGate;
									goto EndCreatePassage;
								}
								else if ((this.grids[y][x] & Grid.Passage) != 0)
								{
									this.grids[y][x] &= ~Grid.RightWall;
									if (x == x2 || (this.grids[y][x] & Grid.LeftWall) != 0 || (intersectRate < 100 && this.rnd.Next(100) >= intersectRate))
									{
										goto EndCreatePassage;
									}
									this.grids[y][x] &= ~Grid.LeftWall;
								}
								else
								{
									this.grids[y][x] |= Grid.HorizontalPassage;
									if (x == x2)
									{
										this.grids[y][x] |= Grid.LeftWall;
									}
								}
							}
						}
						else
						{
							if (j == 1)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y][x] |= Grid.Passage | Grid.RightGate;
									this.grids[y][x + 1] |= Grid.LeftGate;
								}
								else
								{
									this.grids[y][x] |= Grid.Passage;
									this.grids[y][x] &= ~Grid.RightWall;
								}
							}
							else
							{
								this.grids[y][x] &= ~Grid.RightWall;
							}
							while (++x <= x2)
							{
								if ((this.grids[y][x] & Grid.Room) != 0)
								{
									this.grids[y][x - 1] |= Grid.RightGate;
									this.grids[y][x] |= Grid.Passage | Grid.LeftGate;
									goto EndCreatePassage;
								}
								else if ((this.grids[y][x] & Grid.Passage) != 0)
								{
									this.grids[y][x] &= ~Grid.LeftWall;
									if (x == x2 || (this.grids[y][x] & Grid.RightWall) != 0 || (intersectRate < 100 && this.rnd.Next(100) >= intersectRate))
									{
										goto EndCreatePassage;
									}
									this.grids[y][x] &= ~Grid.RightWall;
								}
								else
								{
									this.grids[y][x] |= Grid.HorizontalPassage;
									if (x == x2)
									{
										this.grids[y][x] |= Grid.RightWall;
									}
								}
							}
						}
					}
				}
			EndCreatePassage:
				roomList.Add(r1);
			}
		}

		private void SetRoomGrids(Rectangle room, bool noOutlet)
		{
			if (noOutlet)
			{
				this.grids[room.Y][room.X] |= Grid.Passage;
			}
			else
			{
				for (int y = room.Top; y <= room.Bottom; ++y)
				{
					for (int x = room.Left; x <= room.Right; ++x)
					{
						this.grids[y][x] |= Grid.Room;
					}
				}
			}
			for (int y = room.Top; y <= room.Bottom; ++y)
			{
				this.grids[y][room.Left] |= Grid.LeftWall;
				this.grids[y][room.Right] |= Grid.RightWall;
			}
			for (int x = room.Left; x <= room.Right; ++x)
			{
				this.grids[room.Top][x] |= Grid.TopWall;
				this.grids[room.Bottom][x] |= Grid.BottomWall;
			}
		}
	}
}
